from ..e_typing import (
    ColorStdOut,
    CanStdOut,
    StdOut,
)
from functools import wraps
from typing import (
    ParamSpec,
    Callable,
    Optional,
    TypeVar,
    TypeVarTuple,
)
from .cprint import (
    Color,
)
import sys
__all__ = [
    'DebugPrint',
    'log',
]
_Args = ParamSpec('_Args')
_Returns = TypeVar('_Returns')


def log(cfunc: Optional[Callable[_Args, _Returns]] = None, /,
        *, debugging: bool = True,
        color: Color = Color.blue,
        show_caller: bool = True,
        show_called: bool = True,
        show_args: bool = True,
        show_returns: bool = True) -> Callable[_Args, _Returns]:
    "develop"
    @wraps(cfunc)
    def decorator(func: Callable[_Args, _Returns]) -> Callable[_Args, _Returns]:
        if not debugging: return func
        @wraps(func)
        def wrapper(*args: _Args.args, **kwargs: _Args.kwargs) -> _Returns:
            show = ''
            flg = False
            if show_caller:
                c_frame = sys._getframe(1)
                _mid = f'{c_frame.f_globals["__name__"]}.'
                _mid += f'{c_frame.f_code.co_name}'
                show += f'Caller: {_mid}'
                del _mid
                flg = True
            if show_called:
                if flg:
                    show += '\n'
                show += f'Called: {func.__name__}'
                flg = True
            if show_args:
                if flg:
                    show += '\n'
                show += f'Args: {args}, {kwargs}'
                flg = True
            if flg:
                color(show)
                del show, flg
            result = func(*args, **kwargs)
            if show_returns:
                color(f'Returns: {result}')
            return result
        return wrapper

    if cfunc is not None:
        return decorator(cfunc)

    return decorator


args = TypeVarTuple('args')


class DebugPrint:
    __slots__ = ('debug', 'nl_num')

    def __init__(self, debug_flg: bool, newline_number: int = 3):
        """デバッグ用ログをデバッグフラグがTrueならcall構文で指定した引数の内容を色違いでコンソール出力してくれるクラス\n
        出力する項目をnewline_number個ごとに改行して出力を見やすくする"""
        self.debug = debug_flg
        self.nl_num = newline_number

    def __call__(self, *args: *args) -> ColorStdOut[*args]:
        """examples:
            >>> debug = DebugPrint(True)
            >>> debug('debug', 'to', 'example')
            (light green)debug: debug, to, example
            >>> debug = DebugPrint(True, newline_number = 2)
            >>> debug('debug', 'to', 'example')
            debug: debug, to,
            \texample"""
        nt = '\n\t'
        spc = ' '
        if self.debug:
            for i in range(len(args)):
                if i == 0:
                    content = args[i]
                else:
                    mid_entry = nt if i % self.nl_num == 0 else spc
                    content = f'{content},{mid_entry}{args[i]}'

            print(f'\033[92mdebug: {content}\033[0m')



from collections import namedtuple
from functools import wraps
from ...memodules.log import cprint
from ..e_typing import (
    ColorStdOut,
    NotIterable,
    Function,
)
from typing import (
    NamedTuple,
    ParamSpec,
    ParamSpecArgs,
    ParamSpecKwargs,
    Callable,
    Generic,
    TypeVar,
    get_args,
    Any,
)
__all__ = [
    # const
    'Color',

    # functions
    # #custom
    'cprint',

    # #presets
    'blue',
    'gray',
    'green',
    'magenta',
    'red',
    'turquoise',
    'yellow',

    # #decoration
    'border',
]
_AT = TypeVar('_AT', bound=Callable[..., Any])


def cprint(*args: _AT,
           color: str,
           undo_when_exit: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Color of Custom`"
def gray(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Gray Color`"
def red(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Red Color`"
def green(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Green Color`"
def yellow(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Yellow Color`"
def blue(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Blue Color`"
def magenta(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Magenta Color`"
def turquoise(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Turquoise Color`"
def border(*values: _AT, default: bool = True) -> ColorStdOut[_AT]:
    "Print to Console in `Border String`"
Color = NamedTuple('Color', [
    ('blue', type(blue)), ('gray', type(gray)), ('green', type(green)), ('magenta', type(magenta)), ('red', type(red)), ('turquoise', type(turquoise)), ('yellow', type(yellow)), ('border', type(border))])
Color = Color(blue, gray, green, magenta, red, turquoise, yellow, border)
Color.__doc__ = "Constant of 'cprint' Function"



"""It provides an extra class that has no function and can only be used for type hints.\n
何の機能も無い、型ヒントだけに使えるエクストラなクラスを提供するやで。"""
from typing import (
    TypeVarTuple,
    Callable,
    Generic,
    TypeVar,
    Any,
)
__all__ = [
    'ColorStdOut',
    'NotIterable',
    'StdOut',
    'SupportIndex',
    'SupportPath',
    'Indexes',
    'Pathes'
]

_AT = TypeVar('_AT')  # T は任意の型を表す型変数
_ET = TypeVar('_ET')
_ATs = TypeVarTuple('_ATs')
_F = TypeVar('_F', bound=Callable[..., Any])


class Function(Callable[_F, _AT]):
    def __getitem__(self, key: _F) -> _AT:
        ...


class NotIterable:
    """Mixin to prevent iteration, without being compatible with Iterable.

    That is, we could do:
        def __iter__(self): raise TypeError()
    But this would make users of this mixin duck type-compatible with
    collections.abc.Iterable - isinstance(foo, Iterable) would be True.

    Luckily, we can instead prevent iteration by setting __iter__ to None, which
    is treated specially.
    """

    __slots__ = ()
    __iter__ = None


class StdOut(Generic[*_ATs]):
    '<TypeHints to indicate that it will be printed to the console>'
    def __str__(self) -> str:
        return '<type hints StdOut>'

    def __repr__(self) -> str:
        return self.__doc__


class ColorStdOut(StdOut[*_ATs], Generic[*_ATs]):
    def __str__(self) -> str:
        return '<type hints ColorStdOut>'


class CanStdOut(Generic[*_ATs]):
    ...


class SupportIndex(Generic[*_ATs]):
    ...

class SupportPath(Generic[*_ATs]):
    ...



"""It provides an extra class that has no function and can only be used for type hints.\n
何の機能も無い、型ヒントだけに使えるエクストラなクラスを提供するやで。"""
from typing import (
    _SpecialForm,
    TypeVarTuple,
    ParamSpec,
    Callable,
    Sequence,
    Generic,
    TypeVar,
)
from sys import (
    stdout,
    __stdout__
)
__all__ = [
    'ColorStdOut',
    'NotIterable',
    'StdOut',
    'SupportIndex',
    'SupportPath',
    'Indexes',
    'Pathes'
]


_AT = TypeVar('_AT')  # T は任意の型を表す型変数
_ATs = TypeVarTuple('_ATs')
_AT_co = TypeVar('_AT_co', covariant=True)


values = TypeVarTuple('values')
default = TypeVar('default', bound=bool)
a: Callable[[*values, default], str]


arg_names = TypeVarTuple('arg_names')
return_type = TypeVar('return_type')


class Function(Callable[[*arg_names], return_type]):
    "developing now"
    ...


class NotIterable:
    """Mixin to prevent iteration, without being compatible with Iterable.
    Iterable との互換性がなく、繰り返しを防止するためにミックスインします。

    That is, we could do:
    つまり、次のことができます:
        def __iter__(self): raise TypeError()
    But this would make users of this mixin duck type-compatible with
    collections.abc.Iterable - isinstance(foo, Iterable) would be True.

    Luckily, we can instead prevent iteration by setting __iter__ to None,
    which is treated specially.
    """

    __slots__ = ()
    __iter__ = None


class StdOut(Generic[*_ATs]):
    'TypeHints to indicate that it will be printed to the console'
    def __str__(self) -> str:
        return '<type hints StdOut>'

    def __repr__(self) -> str:
        return f'<{self.__doc__}>'


class ColorStdOut(StdOut[*_ATs], Generic[*_ATs]):
    'TypeHints to indicate that it will be printed of ColorString to the console'
    def __str__(self) -> str:
        return '<type hints ColorStdOut>'


class CanStdOut(Generic[*_ATs]):
    ...


class SupportIndex(Generic[*_ATs]):
    ...


class SupportPath(Generic[*_ATs]):
    ...
